import typing
from typing import Optional

from patronus.evals import EvaluationResult
from patronus.api import api_types

import pydantic


class TaskResult(pydantic.BaseModel):
    output: Optional[str] = None
    metadata: Optional[dict[str, typing.Any]] = None
    tags: Optional[dict[str, str]] = None


MaybeEvaluationResult = typing.Union[EvaluationResult, api_types.EvaluationResult, None]


class EvalsMap(dict):
    def __contains__(self, item) -> bool:
        item = self._key(item)
        return super().__contains__(item)

    def __getitem__(self, item) -> MaybeEvaluationResult:
        item = self._key(item)
        return super().__getitem__(item)

    def __setitem__(self, key: str, value: MaybeEvaluationResult):
        key = self._key(key)
        return super().__setitem__(key, value)

    @staticmethod
    def _key(item):
        if isinstance(item, str):
            return item
        if hasattr(item, "canonical_name"):
            return item.canonical_name
        return item


class _EvalParent(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    task: Optional[TaskResult]
    evals: typing.Optional[EvalsMap]
    parent: typing.Optional["_EvalParent"]

    def find_eval_result(self, evaluator_or_name) -> typing.Union[api_types.EvaluationResult, EvaluationResult, None]:
        if not self.evals and self.parent:
            return self.parent.find_eval_result(evaluator_or_name)
        if evaluator_or_name in self.evals:
            return self.evals[evaluator_or_name]
        return None


_EvalParent.model_rebuild()

EvalParent = typing.Optional[_EvalParent]



from typing import Optional
import warnings


def merge_tags(tags: Optional[dict], new_tags: Optional[dict], experiment_tags: Optional[dict]) -> dict:
    tags = tags or {}
    new_tags = new_tags or {}
    experiment_tags = experiment_tags or {}

    tags = {**tags, **new_tags}
    common_keys = set(tags.keys()) & set(experiment_tags.keys())
    diff = {key for key in common_keys if tags[key] != experiment_tags[key]}
    if diff:
        warnings.warn(f"Overriding experiment tags is not allowed. Tried to override tag: {list(diff)!r}")
    return {**tags, **experiment_tags}



import dataclasses
import logging
from opentelemetry import trace
from typing import Optional
from typing import TYPE_CHECKING

from patronus.context.context_utils import ContextObject
from patronus.exceptions import UninitializedError

if TYPE_CHECKING:
    from patronus.evals.exporter import BatchEvaluationExporter
    from patronus.tracing.logger import Logger as PatLogger
    from patronus.api import PatronusAPIClient
    from opentelemetry.sdk.trace import TracerProvider


@dataclasses.dataclass(frozen=True)
class PatronusScope:
    project_name: Optional[str]
    app: Optional[str]
    experiment_id: Optional[str]


@dataclasses.dataclass(frozen=True)
class PatronusContext:
    scope: PatronusScope
    logger: logging.Logger
    pat_logger: "PatLogger"
    tracer_provider: "TracerProvider"
    tracer: trace.Tracer
    api_client: "PatronusAPIClient"
    exporter: "BatchEvaluationExporter"


_CTX_PAT = ContextObject[PatronusContext]("ctx.pat")


def set_global_patronus_context(ctx: PatronusContext):
    _CTX_PAT.set_global(ctx)


def get_current_context_or_none() -> Optional[PatronusContext]:
    return _CTX_PAT.get()


def get_current_context() -> PatronusContext:
    ctx = get_current_context_or_none()
    if ctx is None:
        raise UninitializedError(
            "No active Patronus context found. Please initialize the library by calling patronus.init()."
        )
    return ctx


def get_logger() -> logging.Logger:
    return get_current_context().logger


def get_logger_or_none() -> Optional[logging.Logger]:
    return (ctx := get_current_context_or_none()) and ctx.logger


def get_pat_logger_or_none() -> Optional["PatLogger"]:
    return (ctx := get_current_context_or_none()) and ctx.pat_logger


def get_tracer() -> trace.Tracer:
    return get_current_context().tracer


def get_tracer_or_none() -> Optional[trace.Tracer]:
    return (ctx := get_current_context_or_none()) and ctx.tracer


def get_api_client() -> "PatronusAPIClient":
    return get_current_context().api_client


def get_api_client_or_none() -> Optional["PatronusAPIClient"]:
    return (ctx := get_current_context_or_none()) and ctx.api_client


def get_exporter() -> "BatchEvaluationExporter":
    return get_current_context().exporter


def get_exporter_or_none() -> Optional["BatchEvaluationExporter"]:
    return (ctx := get_current_context_or_none()) and ctx.exporter


def get_scope() -> PatronusScope:
    return get_current_context().scope


def get_scope_or_none() -> Optional[PatronusScope]:
    return (ctx := get_current_context_or_none()) and ctx.scope



import atexit
import contextvars
import functools
import typing
from multiprocessing.pool import ThreadPool, AsyncResult, ApplyResult
from typing import Union, Optional

from .container import EvaluationContainer
from ..evals import StructuredEvaluator, EvaluationResult, RemoteEvaluator
from ..evals.evaluators import bundled_eval

T = typing.TypeVar("T")

_EvaluatorID = str
_Criteria = str


class EvaluatorDict(typing.TypedDict, total=False):
    evaluator_id: str
    criteria: Optional[_Criteria]


Evaluator = Union[
    StructuredEvaluator,
    EvaluatorDict,
    tuple[_EvaluatorID, _Criteria],
]


# Type Hint for multiprocessing.AsyncResult
# The stdlib doesn't provide a proper generic hint, so we create a monkey-patch-style type hint.
class TypedAsyncResult(typing.Generic[T]):
    def get(self, timeout: Optional[float] = None) -> T: ...
    def wait(self, timeout: Optional[float] = None) -> None: ...
    def ready(self) -> bool: ...
    def successful(self) -> bool: ...


def _into_thread_run_fn(eval_fn, *args, **kwargs) -> typing.Callable[[...], typing.Any]:
    """
    Prepare a function to run in a thread.
    This function make sure that contextvars are propagated as it is necessary for proper
    evaluation and telemetry tracing across thread boundary.
    """
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, eval_fn, *args, **kwargs)
    return func_call


class Patronus:
    def __init__(self, workers: int = 10, shutdown_on_exit: bool = True):
        self._worker_pool = ThreadPool(workers)
        self._supervisor_pool = ThreadPool(workers)

        self._at_exit_handler = None
        if shutdown_on_exit:
            self._at_exit_handler = atexit.register(self.close)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _submit_to_pool(self, callables) -> list[ApplyResult]:
        return [self._worker_pool.apply_async(func=cb) for cb in callables]

    def _process_batch(self, callables, *, return_exceptions: bool) -> list[Union[EvaluationResult, None, Exception]]:
        results = self._submit_to_pool(callables)

        def handle_result(res: AsyncResult):
            try:
                return res.get()
            except Exception as e:
                if return_exceptions:
                    return e
                else:
                    raise e

        return [handle_result(res) for res in results]

    def _map_evaluators(self, evs: list[Evaluator]):
        def _into(ev: Evaluator):
            if isinstance(ev, tuple):
                return RemoteEvaluator(ev[0], ev[1])
            if isinstance(ev, dict):
                return RemoteEvaluator(ev["evaluator_id"], ev["criteria"])
            return ev

        return [_into(e) for e in evs]

    def evaluate(
        self,
        evaluators: typing.Union[list[Evaluator], Evaluator],
        *,
        system_prompt: typing.Optional[str] = None,
        task_context: typing.Union[list[str], str, None] = None,
        task_input: typing.Optional[str] = None,
        task_output: typing.Optional[str] = None,
        gold_answer: typing.Optional[str] = None,
        task_metadata: typing.Optional[dict[str, typing.Any]] = None,
        return_exceptions: bool = False,
    ) -> EvaluationContainer:
        if not isinstance(evaluators, list):
            evaluators = [evaluators]
        evaluators = self._map_evaluators(evaluators)

        with bundled_eval():
            callables = [
                _into_thread_run_fn(
                    ev.evaluate,
                    system_prompt=system_prompt,
                    task_context=task_context,
                    task_input=task_input,
                    task_output=task_output,
                    gold_answer=gold_answer,
                    task_metadata=task_metadata,
                )
                for ev in evaluators
            ]
            results = self._process_batch(callables, return_exceptions=return_exceptions)
            return EvaluationContainer(results)

    def evaluate_bg(
        self,
        evaluators: list[StructuredEvaluator],
        *,
        system_prompt: typing.Optional[str] = None,
        task_context: typing.Union[list[str], str, None] = None,
        task_input: typing.Optional[str] = None,
        task_output: typing.Optional[str] = None,
        gold_answer: typing.Optional[str] = None,
        task_metadata: typing.Optional[dict[str, typing.Any]] = None,
    ) -> TypedAsyncResult[EvaluationContainer]:
        def _run():
            with bundled_eval():
                callables = [
                    _into_thread_run_fn(
                        ev.evaluate,
                        system_prompt=system_prompt,
                        task_context=task_context,
                        task_input=task_input,
                        task_output=task_output,
                        gold_answer=gold_answer,
                        task_metadata=task_metadata,
                    )
                    for ev in evaluators
                ]
                results = self._process_batch(callables, return_exceptions=True)
                return EvaluationContainer(results)

        return typing.cast(
            TypedAsyncResult[EvaluationContainer], self._supervisor_pool.apply_async(_into_thread_run_fn(_run))
        )

    def close(self):
        self._close()
        if self._at_exit_handler:
            atexit.unregister(self._at_exit_handler)

    def _close(self):
        self._supervisor_pool.close()
        self._supervisor_pool.join()
        self._worker_pool.close()
        self._worker_pool.join()



import warnings

from typing import Optional

import httpx

from . import config
from . import context
from .api.api_client import PatronusAPIClient
from .evals.exporter import BatchEvaluationExporter
from .tracing.logger import create_logger, create_patronus_logger
from .tracing.tracer import create_tracer_provider


def init(
    # Initialize SDK with project name.
    # If value is not provided, then it's loaded from the configuration file.
    # Defaults to "Global" if neither is provided.
    project_name: Optional[str] = None,
    # Initialize SDK with an app.
    # If value is not provided, then it's loaded from the configuration file.
    # Defaults to "default" if neither is provided.
    app: Optional[str] = None,
    api_url: Optional[str] = None,
    otel_endpoint: Optional[str] = None,
    api_key: Optional[str] = None,
    **kwargs,
):
    if api_url != config.DEFAULT_API_URL and otel_endpoint == config.DEFAULT_OTEL_ENDPOINT:
        raise ValueError(
            "'api_url' is set to non-default value, "
            "but 'otel_endpoint' is a default. Change 'otel_endpoint' to point to the same environment as 'api_url'"
        )

    cfg = config.config()
    ctx = build_context(
        project_name=project_name or cfg.project_name,
        app=app or cfg.app,
        experiment_id=None,
        api_url=api_url or cfg.api_url,
        otel_endpoint=otel_endpoint or cfg.otel_endpoint,
        api_key=api_key or cfg.api_key,
        timeout_s=cfg.timeout_s,
        **kwargs,
    )
    context.set_global_patronus_context(ctx)


def build_context(
    project_name: str,
    app: Optional[str],
    experiment_id: Optional[str],
    api_url: Optional[str],
    otel_endpoint: str,
    api_key: str,
    client_http: Optional[httpx.Client] = None,
    client_http_async: Optional[httpx.AsyncClient] = None,
    timeout_s: int = 60,
    **kwargs,
) -> context.PatronusContext:
    if client_http is None:
        client_http = httpx.Client(timeout=timeout_s)
    if client_http_async is None:
        client_http_async = httpx.AsyncClient(timeout=timeout_s)
    scope = context.PatronusScope(
        project_name=project_name,
        app=app,
        experiment_id=experiment_id,
    )
    api = PatronusAPIClient(
        client_http_async=client_http_async,
        client_http=client_http,
        base_url=api_url,
        api_key=api_key,
    )
    std_logger = create_logger(
        scope=scope,
        exporter_endpoint=otel_endpoint,
        api_key=api_key,
    )
    eval_logger = create_patronus_logger(
        scope=scope,
        exporter_endpoint=otel_endpoint,
        api_key=api_key,
    )
    tracer_provider = create_tracer_provider(
        exporter_endpoint=otel_endpoint,
        api_key=api_key,
        scope=scope,
    )
    if integrations := kwargs.get("integrations"):
        if not isinstance(integrations, list):
            integrations = [integrations]

        try:
            from opentelemetry.instrumentation.instrumentor import BaseInstrumentor

            for integration in integrations:
                if isinstance(integration, BaseInstrumentor):
                    integration.instrument(tracer_provider=tracer_provider)
                else:
                    warnings.warn(f"Integration {integration} not recognized.")
        except ImportError:
            warnings.warn("Opentelemetry instrumentation is not installed. Ignoring integrations.")

    tracer = tracer_provider.get_tracer("patronus.sdk")

    eval_exporter = BatchEvaluationExporter(client=api)
    return context.PatronusContext(
        scope=scope,
        logger=std_logger,
        pat_logger=eval_logger,
        tracer_provider=tracer_provider,
        tracer=tracer,
        api_client=api,
        exporter=eval_exporter,
    )



from enum import Enum
from typing import Optional


class Attributes(str, Enum):
    log_id = "pat.log.id"
    log_type = "pat.log.type"
    project_name = "pat.project.name"
    app = "pat.app"
    experiment_id = "pat.experiment.id"
    evaluator_id = "pat.evaluator.id"
    evaluator_criteria = "pat.evaluator.criteria"


class GenAIAttributes(str, Enum):
    operation_name = "gen_ai.operation.name"


class OperationNames(str, Enum):
    eval = "eval"
    task = "task"


class LogTypes(str, Enum):
    # Eval log type is emitted by evaluators, they contain evaluation data
    eval = "eval"
    # Trace log type is emitted by traced() decorator.
    trace = "trace"
    # User log type is emitted any time user uses the logger directly.
    user = "user"


def format_service_name(project_name: str, app: Optional[str] = None, experiment_id: Optional[str] = None) -> str:
    service_name = f"{project_name}/"
    if experiment_id:
        service_name += f"ex:{experiment_id}"
    else:
        service_name += f"app:{app}"
    return service_name



import dataclasses
from typing import Union, Generator

from patronus.evals import EvaluationResult
from patronus.exceptions import MultiException


@dataclasses.dataclass
class EvaluationContainer:
    results: list[Union[EvaluationResult, None, Exception]]

    def has_exception(self) -> bool:
        return any(isinstance(r, Exception) for r in self.results)

    def raise_on_exception(self) -> None:
        if not self.has_exception():
            return None
        exceptions = list(r for r in self.results if isinstance(r, Exception))
        if len(exceptions) == 1:
            raise exceptions[0]
        raise MultiException(exceptions)

    def all_succeeded(self, ignore_exceptions: bool = False) -> bool:
        """
        Check if all evaluations that were actually evaluated passed.

        Evaluations are only considered if they:
        - Have a non-None pass_ flag set
        - Are not None (skipped)
        - Are not exceptions (unless ignore_exceptions=True)

        Note: Returns True if no evaluations met the above criteria (empty case).
        """
        for r in self.results:
            if isinstance(r, Exception) and not ignore_exceptions:
                self.raise_on_exception()
            if r is not None and r.pass_ is False:
                return False
        return True

    def any_failed(self, ignore_exceptions: bool = False) -> bool:
        """
        Check if any evaluation that was actually evaluated failed.

        Evaluations are only considered if they:
        - Have a non-None pass_ flag set
        - Are not None (skipped)
        - Are not exceptions (unless ignore_exceptions=True)

        Note: Returns False if no evaluations met the above criteria (empty case).
        """
        for r in self.results:
            if isinstance(r, Exception) and not ignore_exceptions:
                self.raise_on_exception()
            if r is not None and r.pass_ is False:
                return True
        return False

    def failed_evaluations(self) -> Generator[EvaluationResult, None, None]:
        return (r for r in self.results if not isinstance(r, (Exception, type(None))) and r.pass_ is False)

    def succeeded_evaluations(self) -> Generator[EvaluationResult, None, None]:
        return (r for r in self.results if not isinstance(r, (Exception, type(None))) and r.pass_ is True)



from asyncio import Task
from concurrent.futures import ThreadPoolExecutor

import asyncio
import collections
import inspect
import typing
from typing import Optional, Union, TypedDict, List

from patronus.evals import bundled_eval
from patronus.evals import StructuredEvaluator
from patronus.evals import AsyncStructuredEvaluator
from patronus.evals import AsyncRemoteEvaluator
from .container import EvaluationContainer

_EvaluatorID = str
_Criteria = str


class EvaluatorDict(TypedDict, total=False):
    evaluator_id: str
    criteria: Optional[_Criteria]


Evaluator = Union[
    StructuredEvaluator,
    AsyncStructuredEvaluator,
    EvaluatorDict,
    tuple[_EvaluatorID, _Criteria],
]


async def with_semaphore(sem: asyncio.Semaphore, coro: typing.Coroutine):
    async with sem:
        return await coro


class AsyncPatronus:
    def __init__(self, max_workers: int = 10):
        self._pending_tasks = collections.deque()
        self._executor = ThreadPoolExecutor(max_workers=max_workers)
        self._semaphore = asyncio.Semaphore(max_workers)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def _map_evaluators(self, evs: list[Evaluator]):
        def _into(ev: Evaluator):
            if isinstance(ev, tuple):
                return AsyncRemoteEvaluator(ev[0], ev[1])
            if isinstance(ev, dict):
                return AsyncRemoteEvaluator(ev["evaluator_id"], ev["criteria"])
            return ev

        return [_into(e) for e in evs]

    async def evaluate(
        self,
        evaluators: Union[List[Evaluator], Evaluator],
        *,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[dict] = None,
        return_exceptions: bool = False,
    ) -> EvaluationContainer:
        singular_eval = not isinstance(evaluators, list)
        if singular_eval:
            evaluators = [evaluators]
        evaluators = self._map_evaluators(evaluators)

        def into_coro(fn, **kwargs):
            if inspect.iscoroutinefunction(fn):
                coro = fn(**kwargs)
            else:
                coro = asyncio.to_thread(fn, **kwargs)
            return with_semaphore(self._semaphore, coro)

        with bundled_eval():
            results = await asyncio.gather(
                *(
                    into_coro(
                        ev.evaluate,
                        system_prompt=system_prompt,
                        task_context=task_context,
                        task_input=task_input,
                        task_output=task_output,
                        gold_answer=gold_answer,
                        task_metadata=task_metadata,
                    )
                    for ev in evaluators
                ),
                return_exceptions=return_exceptions,
            )
        return EvaluationContainer(results)

    def evaluate_bg(
        self,
        evaluators: Union[List[Evaluator], Evaluator],
        *,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[dict] = None,
    ) -> Task[EvaluationContainer]:
        loop = asyncio.get_running_loop()
        task = loop.create_task(
            self.evaluate(
                evaluators=evaluators,
                system_prompt=system_prompt,
                task_context=task_context,
                task_input=task_input,
                task_output=task_output,
                gold_answer=gold_answer,
                task_metadata=task_metadata,
                return_exceptions=True,
            ),
            name="evaluate_bg",
        )
        self._pending_tasks.append(task)
        task.add_done_callback(self._consume_tasks)
        return task

    def _consume_tasks(self, task):
        while len(self._pending_tasks) > 0:
            task: Task = self._pending_tasks[0]
            if task.done():
                self._pending_tasks.popleft()
            else:
                return

    async def close(self):
        while len(self._pending_tasks) != 0:
            await self._pending_tasks.popleft()



from .api.api_types import EvaluateRequest as EvaluateRequest
from .context import get_api_client as get_api_client
from .context import get_logger as get_logger

from .evals import AsyncStructuredEvaluator as AsyncStructuredEvaluator
from .evals import EvaluationResult as EvaluationResult
from .evals import Evaluator as Evaluator
from .evals import StructuredEvaluator as StructuredEvaluator
from .evals import evaluator as evaluator
from .evals import RemoteEvaluator as RemoteEvaluator


from .init import init as init
from .pat_client import AsyncPatronus as AsyncPatronus
from .pat_client import Patronus as Patronus


from .tracing.decorators import traced as traced
from .tracing.decorators import start_span as start_span



import dataclasses
import importlib.metadata
import logging
import platform
import typing
from functools import lru_cache
from typing import Optional

import httpx
import pydantic

from patronus.exceptions import WithStackTraceMixin

log = logging.getLogger("patronus.core")


class APIError(Exception):
    response: httpx.Response

    def __init__(self, message: Optional[str] = None, response: Optional[httpx.Response] = None):
        self.response = response
        super().__init__(message)


class ClientError(APIError):
    def __init__(self, response: httpx.Response):
        self.response = response
        super().__init__(f"Client error '{response.status_code}' for url '{response.request.url}: {response.text}'")


class UnrecoverableAPIError(APIError):
    def __init__(self, message: str, response: httpx.Response):
        message = f"{self.__class__.__name__}: {message}: response content: {response.text}"
        super().__init__(message, response)


class RPMLimitError(APIError):
    def __init__(self, limit: int, wait_for_s: Optional[float], response: httpx.Response):
        self.limit = limit
        self.wait_for_s = wait_for_s
        super().__init__(
            f"RPMLimitError: Request per minute limit hit (quota limit: {self.limit})",
            response=response,
        )


class RetryError(WithStackTraceMixin, Exception):
    def __init__(self, attempt: int, out_of: int, origin: Exception, stack_trace: str):
        self.stack_trace = stack_trace
        super().__init__(f"RetryError: execution failed after {attempt}/{out_of} attempts: {origin}")


R = typing.TypeVar("R", bound=pydantic.BaseModel)


@dataclasses.dataclass
class CallResponse(typing.Generic[R]):
    data: Optional[R]
    response: httpx.Response

    def raise_for_status(self):
        if self.response.status_code in (400, 422):
            raise ClientError(self.response)
        self.response.raise_for_status()


class BaseAPIClient:
    version: str
    http: httpx.AsyncClient
    http_sync: httpx.Client
    base_url: str = None
    api_key: str = None

    def __init__(
        self,
        *,
        client_http_async: httpx.AsyncClient,
        client_http: httpx.Client,
        base_url: str,
        api_key: str,
    ):
        self.version = importlib.metadata.version("patronus")
        self.http = client_http_async
        self.http_sync = client_http
        self.base_url = base_url.rstrip("/")
        self.api_key = api_key

    def set_target(self, base_url: str, api_key: str):
        self.base_url = base_url.rstrip("/")
        self.api_key = api_key

    async def call(
        self,
        method: str,
        path: str,
        *,
        params: Optional[dict[str, typing.Any]] = None,
        body: Optional[pydantic.BaseModel] = None,
        response_cls: Optional[type[R]],
    ) -> CallResponse[R]:
        content = None
        if body is not None:
            content = body.model_dump_json(by_alias=True)

        url = self._url(path)
        log.debug(f"Sending HTTP request {url!r}")
        response = await self.http.request(
            method,
            url,
            params=params,
            content=content,
            headers=self.headers(),
        )
        return self._call_process_resp(url, response, response_cls)

    def call_sync(
        self,
        method: str,
        path: str,
        *,
        params: Optional[dict[str, typing.Any]] = None,
        body: Optional[pydantic.BaseModel] = None,
        response_cls: Optional[type[R]],
    ) -> CallResponse[R]:
        content = None
        if body is not None:
            content = body.model_dump_json(by_alias=True)

        url = self._url(path)
        log.debug(f"Sending HTTP request {url!r}")
        response = self.http_sync.request(
            method,
            url,
            params=params,
            content=content,
            headers=self.headers(),
        )
        return self._call_process_resp(url, response, response_cls)

    @staticmethod
    def _call_process_resp(url: str, response: httpx.Response, response_cls: Optional[type[R]]) -> CallResponse[R]:
        log.debug(f"Received HTTP response {url!r} {response.status_code}")

        data = None
        if response.is_error:
            log.debug(f"Response with error: content: {response.text}")
        if response.is_success and response_cls is not None:
            data = response_cls.model_validate_json(response.content)

        return CallResponse(
            data=data,
            response=response,
        )

    def _url(self, path: str):
        assert self.base_url, "BaseAPIClient: base_url must be set"
        return f"{self.base_url}{path}"

    def headers(self) -> typing.Dict[str, str]:
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": self.user_agent,
            **self.auth_headers(),
            **self.platform_headers(),
        }

    @property
    def user_agent(self) -> str:
        return f"PatronusAI/Python {self.version}"

    def auth_headers(self) -> typing.Dict[str, str]:
        assert self.api_key, "BaseAPIClient.auth_headers(): api_key must be set"
        return {
            "x-api-key": self.api_key,
        }

    @lru_cache(maxsize=None)
    def platform_headers(self) -> typing.Dict[str, str]:
        return {
            "X-Package-Version": self.version,
            "X-Platform-OS": get_platform_os(),
            "X-Platform-Arch": get_platform_arch(),
            "X-Runtime": get_python_runtime(),
            "X-Runtime-version": get_python_version(),
        }


# get_platform_os is based on implementation found in Open AI SDK
# https://github.com/openai/openai-python/blob/631a2a7156299351874f37c4769308a104ce19ed/src/openai/_base_client.py#L1933-L1972
def get_platform_os() -> str:
    try:
        system = platform.system().lower()
        platform_name = platform.platform().lower()
    except Exception:
        return "Unknown"

    if "iphone" in platform_name or "ipad" in platform_name:
        return "iOS"

    if system == "darwin":
        return "MacOS"

    if system == "windows":
        return "Windows"

    if "android" in platform_name:
        return "Android"

    if system == "linux":
        return "Linux"

    if platform_name:
        return f"Other:{platform_name}"

    return "Unknown"


def get_platform_arch():
    try:
        python_bitness, _ = platform.architecture()
        machine = platform.machine().lower()
    except Exception:
        return "Unknown"

    if machine in ("arm64", "aarch64"):
        return "arm64"

    if machine == "arm":
        return "arm"

    if machine == "x86_64":
        return "x64"

    if python_bitness == "32bit":
        return "x32"

    return "Unknown"


def get_python_runtime() -> str:
    try:
        return platform.python_implementation()
    except Exception:
        return "Unknown"


def get_python_version() -> str:
    try:
        return platform.python_version()
    except Exception:
        return "Unknown"



import functools
import typing
from typing import Optional

import pydantic
import pydantic_settings

DEFAULT_API_URL = "https://api.patronus.ai"
DEFAULT_OTEL_ENDPOINT = "https://otel.patronus.ai:4317"
DEFAULT_UI_URL = "https://app.patronus.ai"

DEFAULT_PROJECT_NAME = "Global"

DEFAULT_APP_NAME = "default"


class Config(pydantic_settings.BaseSettings):
    model_config = pydantic_settings.SettingsConfigDict(env_prefix="patronus_", yaml_file="patronus.yaml")

    api_key: Optional[str] = pydantic.Field(default=None)
    api_url: str = pydantic.Field(default=DEFAULT_API_URL)
    otel_endpoint: str = pydantic.Field(default=DEFAULT_OTEL_ENDPOINT)
    ui_url: str = pydantic.Field(default=DEFAULT_UI_URL)

    timeout_s: int = pydantic.Field(
        default=300,
        description=(
            "Timeout for http client connecting to Patronus APIs. "
            "This includes calls for remote evaluations which may take more time "
            "than most API calls. Because of that the timeout should be greater than usual."
        ),
    )

    project_name: str = pydantic.Field(default=DEFAULT_PROJECT_NAME)
    app: str = pydantic.Field(default=DEFAULT_APP_NAME)

    @pydantic.model_validator(mode="after")
    def validate_otel_endpoint(self) -> "Config":
        if self.api_url != DEFAULT_API_URL and self.otel_endpoint == DEFAULT_OTEL_ENDPOINT:
            raise ValueError(
                "configuration error: 'api_url' is set to non-default value, "
                "but 'otel_endpoint' is a default. Change 'otel_endpoint' to point to the same environment as 'api_url'"
            )
        return self

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: typing.Type[pydantic_settings.BaseSettings],
        init_settings: pydantic_settings.PydanticBaseSettingsSource,
        env_settings: pydantic_settings.PydanticBaseSettingsSource,
        dotenv_settings: pydantic_settings.PydanticBaseSettingsSource,
        file_secret_settings: pydantic_settings.PydanticBaseSettingsSource,
    ) -> typing.Tuple[pydantic_settings.PydanticBaseSettingsSource, ...]:
        return (
            pydantic_settings.YamlConfigSettingsSource(settings_cls),
            pydantic_settings.EnvSettingsSource(settings_cls),
        )


@functools.lru_cache()
def config() -> Config:
    cfg = Config()
    return cfg



import datetime
import pydantic
import textwrap
from typing import Any, Optional


class EvaluationResult(pydantic.BaseModel):
    score: Optional[float] = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
        Score of the evaluation. Can be any numerical value, though typically ranges from 0 to 1,
        where 1 represents the best possible score.
        """
        ),
    )
    # Whether the evaluation is considered to pass or fail
    pass_: Optional[bool] = None
    # Text output of the evaluation.
    # Usually used for discrete human-readable category evaluation or as a label for score value.
    text_output: Optional[str] = None
    # Arbitrary json-serializable metadata about evaluation.
    metadata: Optional[dict[str, Any]] = None
    # Human-readable explanation of the evaluation.
    explanation: Optional[str] = None
    # Key-value pair metadata
    tags: Optional[dict[str, str]] = None
    # ID of the dataset associated with evaluated sample
    dataset_id: Optional[str] = None
    # ID of the sample in a dataset associated with evaluated sample
    dataset_sample_id: Optional[str] = None
    # Duration of the evaluation.
    # In case value is not set, @evaluator decorator and Evaluator classes will set this value automatically.
    evaluation_duration: Optional[datetime.timedelta] = None
    # Duration of the evaluation explanation.
    explanation_duration: Optional[datetime.timedelta] = None



import contextlib
import functools
import inspect
import typing
from typing import Optional

from opentelemetry.util.types import Attributes
from opentelemetry._logs import SeverityNumber

from patronus.tracing.attributes import LogTypes
from patronus.tracing.logger import Logger
from patronus import context


@contextlib.contextmanager
def start_span(name: str, *, record_exception: bool = True, attributes: Optional[Attributes] = None):
    ctx = context.get_current_context_or_none()
    if ctx is None:
        yield
        return
    with ctx.tracer.start_as_current_span(
        name,
        record_exception=record_exception,
        attributes=attributes,
    ) as span:
        yield span


def traced(
    # Give name for the traced span. Defaults to a function name if not provided.
    span_name: Optional[str] = None,
    *,
    # Whether to log function arguments.
    log_args: bool = True,
    # Whether to log function output.
    log_results: bool = True,
    # Whether to log an exception if one was raised.
    log_exceptions: bool = True,
    # Whether to prevent a log message to be created.
    disable_log: bool = False,
    attributes: Attributes = None,
    **kwargs,
):
    def decorator(func):
        name = span_name or func.__qualname__
        sig = inspect.signature(func)
        record_exception = not disable_log and log_exceptions

        def log_call(logger: Logger, fn_args: typing.Any, fn_kwargs: typing.Any, ret: typing.Any, exc: Exception):
            if disable_log:
                return

            severity = SeverityNumber.INFO
            body = {"function.name": name}
            if log_args:
                bound_args = sig.bind(*fn_args, **fn_kwargs)
                body["function.arguments"] = {**bound_args.arguments, **bound_args.arguments}
            if log_results is not None and exc is None:
                body["function.output"] = ret
            if log_exceptions and exc is not None:
                module = type(exc).__module__
                qualname = type(exc).__qualname__
                exception_type = f"{module}.{qualname}" if module and module != "builtins" else qualname
                body["exception.type"] = exception_type
                body["exception.message"] = str(exc)
                severity = SeverityNumber.ERROR
            logger.log(body, log_type=LogTypes.trace, severity=severity)

        @functools.wraps(func)
        def wrapper_sync(*f_args, **f_kwargs):
            ctx = context.get_current_context_or_none()
            if ctx is None:
                return func(*f_args, **f_kwargs)

            exc = None
            ret = None
            with ctx.tracer.start_as_current_span(name, record_exception=record_exception, attributes=attributes):
                try:
                    ret = func(*f_args, **f_kwargs)
                except Exception as e:
                    exc = e
                    raise exc
                finally:
                    log_call(ctx.pat_logger, f_args, f_kwargs, ret, exc)

                return ret

        @functools.wraps(func)
        async def wrapper_async(*f_args, **f_kwargs):
            ctx = context.get_current_context_or_none()
            if ctx is None:
                return await func(*f_args, **f_kwargs)

            exc = None
            ret = None
            with ctx.tracer.start_as_current_span(name, record_exception=record_exception, attributes=attributes):
                try:
                    ret = await func(*f_args, **f_kwargs)
                except Exception as e:
                    exc = e
                    raise exc
                finally:
                    log_call(ctx.pat_logger, f_args, f_kwargs, ret, exc)

                return ret

        if inspect.iscoroutinefunction(func):
            wrapper_async._pat_traced = True
            return wrapper_async
        else:
            wrapper_async._pat_traced = True
            return wrapper_sync

    return decorator



from .adapters import BaseEvaluatorAdapter as BaseEvaluatorAdapter
from .adapters import EvaluatorAdapter as EvaluatorAdapter
from .adapters import StructuredEvaluatorAdapter as StructuredEvaluatorAdapter
from .adapters import FuncEvaluatorAdapter as FuncEvaluatorAdapter

from .experiment import run_experiment as run_experiment
from .experiment import Experiment as Experiment

from .types import TaskResult as TaskResult
from .types import EvalParent as EvalParent

from ..datasets import Dataset as Dataset
from ..datasets import Row as Row



import asyncio
import httpx
import inspect
import os
import pandas as pd
import time
import typing
from typing import Optional, Any, Union

import typing_extensions as te

from patronus import context, datasets
from patronus.api import PatronusAPIClient, api_types
from patronus.context import get_tracer
from patronus.datasets import Dataset, DatasetLoader
from patronus.evals import StructuredEvaluator, AsyncStructuredEvaluator, bundled_eval, EvaluationResult
from patronus.evals.context import evaluation_attributes
from patronus.experiments.adapters import BaseEvaluatorAdapter, StructuredEvaluatorAdapter
from patronus.experiments.async_utils import run_until_complete
from patronus.experiments.reporter import Reporter
from patronus.experiments.tqdm import AsyncTQDMWithHandle
from patronus.experiments.types import EvalParent, TaskResult, _EvalParent, EvalsMap
from patronus.tracing import traced
from patronus.utils import merge_tags

# TODO Type hint them
Tags = dict[str, str]

T = typing.TypeVar("T")


class TaskProtocol(typing.Protocol[T]):
    def __call__(self, *, row: datasets.Row, parent: EvalParent, tags: Tags) -> T: ...


Task = Union[
    # Synchronous task signature
    TaskProtocol[Union[TaskResult, str, None]],
    # Asynchronous task signature
    TaskProtocol[typing.Awaitable[Union[TaskResult, str, None]]],
]

ExperimentDataset = Union[
    Dataset,
    DatasetLoader,
    list[dict[str, Any]],
    tuple[dict[str, Any], ...],
    pd.DataFrame,
    typing.Awaitable,
    typing.Callable[[], typing.Awaitable],
]

AdaptableEvaluators = Union[StructuredEvaluator, AsyncStructuredEvaluator, BaseEvaluatorAdapter]


class ChainLink(typing.TypedDict):
    task: Optional[Task]
    evaluators: list[AdaptableEvaluators]


class _ChainLink(typing.TypedDict):
    task: Optional[Task]
    evaluators: list[BaseEvaluatorAdapter]


def run_experiment(
    dataset: ExperimentDataset,
    task: Optional[Task] = None,
    evaluators: Optional[list[AdaptableEvaluators]] = None,
    chain: Optional[list[ChainLink]] = None,
    tags: Optional[Tags] = None,
    max_concurrency: int = 10,
    project_name: Optional[str] = None,
    experiment_name: Optional[str] = None,
    api_key: Optional[str] = None,
    **kwargs,
) -> "Experiment":
    """
    Create and run an experiment.

    This function creates an experiment with the specified configuration and runs it to completion.
    The execution handling is context-aware:

    - When called from an asynchronous context (with a running event loop), it returns an
      awaitable that must be awaited.
    - When called from a synchronous context (no running event loop), it blocks until the
      experiment completes and returns the Experiment object.


    Returns
    -------
    Union[Experiment, Awaitable[Experiment]]
        In a synchronous context: the completed Experiment object.
        In an asynchronous context: an awaitable that resolves to the Experiment object.

    Examples:
        Synchronous execution:
            experiment = run_experiment(dataset, task=some_task)
            # Blocks until the experiment finishes.

        Asynchronous execution (e.g., in a Jupyter Notebook):
            experiment = await run_experiment(dataset, task=some_task)
            # Must be awaited within an async function or event loop.

    Notes:
        For manual control of the event loop, you can create and run the experiment as follows:
            experiment = await Experiment.create(...)
            await experiment.run()

    """

    async def _run_experiment() -> Union[Experiment, typing.Awaitable[Experiment]]:
        ex = await Experiment.create(
            dataset=dataset,
            task=task,
            evaluators=evaluators,
            chain=chain,
            tags=tags,
            max_concurrency=max_concurrency,
            project_name=project_name,
            experiment_name=experiment_name,
            api_key=api_key,
            **kwargs,
        )
        return await ex.run()

    return run_until_complete(_run_experiment())


class Experiment:
    project: Optional[api_types.Project]
    experiment: Optional[api_types.Experiment]
    tags: dict[str, str]
    # dataset is transformed raw dataset that is used by the experiment.
    dataset: Optional[Dataset]

    _project_name: Optional[str]
    _experiment_name: Optional[str]
    # _raw_dataset is a raw object passed to the constructor. It may be unset after Experiment is prepared.
    _raw_dataset: Optional[ExperimentDataset]

    _chain: list[_ChainLink]
    _started: bool

    _sem_tasks: asyncio.Semaphore
    _sem_evals: asyncio.Semaphore

    _api_key: Optional[str]

    _ctx: Optional[context.PatronusContext] = None

    def __init__(
        self,
        *,
        dataset: typing.Any,
        task: Optional[Task] = None,
        evaluators: Optional[list[AdaptableEvaluators]] = None,
        chain: Optional[list[ChainLink]] = None,
        tags: Optional[dict[str, str]] = None,
        max_concurrency: int = 10,
        project_name: Optional[str] = None,
        experiment_name: Optional[str] = None,
        # TODO
        # ...
        api_key: Optional[str] = None,
        **kwargs,
    ):
        if chain and evaluators:
            raise ValueError("Cannot specify both chain and evaluators")

        self._raw_dataset = dataset

        if not chain:
            chain = [{"task": task, "evaluators": evaluators}]
        self._chain = [
            {"task": _trace_task(link["task"]), "evaluators": _adapt_evaluators(link["evaluators"])} for link in chain
        ]
        self._started = False
        self._finished = False

        self._project_name = project_name
        self.project = None

        self._experiment_name = experiment_name
        self.experiment = None

        self.tags = tags or {}

        self.max_concurrency = max_concurrency

        self._api_key = api_key
        self._prepared = False

        self.reporter = Reporter()

        self._integrations = kwargs.get("integrations")

    @classmethod
    async def create(
        cls,
        dataset: ExperimentDataset,
        task: Optional[Task] = None,
        evaluators: Optional[list[AdaptableEvaluators]] = None,
        chain: Optional[list[ChainLink]] = None,
        tags: Optional[Tags] = None,
        max_concurrency: int = 10,
        project_name: Optional[str] = None,
        experiment_name: Optional[str] = None,
        api_key: Optional[str] = None,
        **kwargs,
    ) -> te.Self:
        ex = cls(
            dataset=dataset,
            task=task,
            evaluators=evaluators,
            chain=chain,
            tags=tags,
            max_concurrency=max_concurrency,
            project_name=project_name,
            experiment_name=experiment_name,
            api_key=api_key,
            **kwargs,
        )
        ex._ctx = await ex._prepare()

        return ex

    async def run(self) -> te.Self:
        if self._started:
            raise RuntimeError("Experiment already started")
        if self._prepared is False:
            raise ValueError(
                "Experiment must be prepared before starting. "
                "Seems that Experiment was not created using Experiment.create() classmethod."
            )
        self._started = True

        with context._CTX_PAT.using(self._ctx):
            await self._run()
            self._finished = True
            self.reporter.summary()

        await asyncio.to_thread(self._ctx.exporter.force_flush)
        await asyncio.to_thread(self._ctx.tracer_provider.force_flush)

        return self

    def to_dataframe(self) -> pd.DataFrame:
        if self._finished is not True:
            raise RuntimeError("Experiment has to be in finished state")
        return self.reporter.to_dataframe()

    def to_csv(self, path_or_buf, **kwargs) -> Optional[str]:
        return self.to_dataframe().to_csv(path_or_buf, **kwargs)

    async def _prepare(self) -> context.PatronusContext:
        # Creating the semaphores here instead of in __init__ to make sure
        # we initialize them in context of an event loop that will run the experiment.
        self._sem_tasks = asyncio.Semaphore(self.max_concurrency)
        self._sem_evals = asyncio.Semaphore(self.max_concurrency)

        from patronus.config import config
        from patronus.init import build_context

        cfg = config()

        client_http = httpx.Client(timeout=cfg.timeout_s)
        client_http_async = httpx.AsyncClient(timeout=cfg.timeout_s)

        api = PatronusAPIClient(
            client_http_async=client_http_async,
            client_http=client_http,
            base_url=cfg.api_url,
            api_key=self._api_key or cfg.api_key,
        )

        dataset = await self._prepare_dataset(self._raw_dataset)
        self._raw_dataset = None
        self.dataset = dataset

        self.project = await self._get_or_create_project(api, self._project_name or cfg.project_name)
        self._project_name = None

        self.experiment = await self._create_experiment(api, self.project.id, self._experiment_name, self.tags)
        self._experiment_name = None

        ctx = build_context(
            project_name=self.project.name,
            app=None,
            experiment_id=self.experiment.id,
            api_url=cfg.api_url,
            otel_endpoint=cfg.otel_endpoint,
            api_key=self._api_key or cfg.api_key,
            client_http=client_http,
            client_http_async=client_http_async,
            integrations=self._integrations,
        )
        self._prepared = True
        return ctx

    async def _run(self):
        title = f"Experiment  {self.project.name}/{self.experiment.name}"
        print("=" * len(title))

        tasks = [
            with_semaphore(self._sem_tasks, self._run_chain(idx, row))
            for idx, row in enumerate(self.dataset.iterrows(), start=1)
        ]

        tqdm = await AsyncTQDMWithHandle.prep_gather(*tasks, desc=title, unit="sample")
        self.reporter.set_tqdm(tqdm)
        await tqdm.gather()

    @classmethod
    async def _prepare_dataset(cls, dataset: Any) -> Dataset:
        if isinstance(dataset, Dataset):
            return dataset
        elif isinstance(dataset, DatasetLoader):
            return await dataset.load()
        elif isinstance(dataset, (list, tuple)):
            return Dataset.from_records(dataset)
        elif inspect.iscoroutine(dataset):
            return await cls._prepare_dataset(await dataset)
        elif inspect.iscoroutinefunction(dataset):
            return await cls._prepare_dataset(await dataset())
        elif callable(dataset):
            return await cls._prepare_dataset(dataset())
        elif isinstance(dataset, pd.DataFrame):
            return Dataset.from_dataframe(dataset)
        else:
            raise ValueError(f"'dataset' passed to the experiment is an unexpected object of type {type(dataset)!r}")

    @staticmethod
    async def _get_or_create_project(api: PatronusAPIClient, project_name: str) -> api_types.Project:
        return await api.create_project(api_types.CreateProjectRequest(name=project_name))

    @staticmethod
    async def _create_experiment(
        api: PatronusAPIClient, project_id: str, experiment_name: str, tags: Tags
    ) -> api_types.Experiment:
        name = generate_experiment_name(experiment_name)
        return await api.create_experiment(
            api_types.CreateExperimentRequest(
                project_id=project_id,
                name=name,
                tags=tags,
            )
        )

    async def _run_chain(self, idx: int, row: datasets.Row):
        tracer = get_tracer()
        span_name = f"Sample {idx}/{self.dataset.df.shape[0]}"
        with tracer.start_as_current_span(span_name):
            parent = None

            for link_idx, eval_link in enumerate(self._chain):
                # TODO come up with better span name
                with tracer.start_as_current_span(f"Stage {link_idx+1}"):
                    task = eval_link["task"]
                    adapted_evaluators: list[BaseEvaluatorAdapter] = eval_link["evaluators"]

                    outgoing_tags = merge_tags({}, row.tags or {}, experiment_tags=self.tags)
                    task_result: Optional[TaskResult] = None

                    if task is not None:
                        try:
                            task_result = await self.execute_task(task, row, parent, outgoing_tags)
                        except Exception as exc:
                            self.reporter.add_task_error(exc, row)
                            return

                        # If task returned None it means the record processing should be skipped
                        if task_result is None:
                            return

                    if task_result is not None and task_result.tags:
                        outgoing_tags = merge_tags(outgoing_tags, task_result.tags, experiment_tags=self.tags)

                    results = await self.evaluate_stage(adapted_evaluators, row, task_result, parent, outgoing_tags)

                    has_eval_errors = False
                    eval_results_map = EvalsMap()

                    for adapter, result in zip(adapted_evaluators, results):
                        if isinstance(result, Exception):
                            has_eval_errors = True
                            self.reporter.add_evaluator_error(result, row, adapter.evaluator_id, adapter.criteria)
                            continue

                        eval_results_map[adapter.canonical_name] = result

                        await self.reporter.add_result(
                            link_idx,
                            task.__name__ if task else None,
                            task_result,
                            adapter.evaluator_id,
                            adapter.criteria,
                            result,
                            row,
                        )

                    if has_eval_errors:
                        return

                    parent = _EvalParent(task=task_result, evals=eval_results_map, parent=parent)

    async def execute_task(self, task, row: datasets.Row, parent: EvalParent, tags: Tags) -> Optional[TaskResult]:
        try:
            if inspect.iscoroutinefunction(task):
                task_result = await task(row=row, parent=parent, tags=tags)
            else:
                # TODO handle with thread pool executor
                task_result = await asyncio.to_thread(task, row=row, parent=parent, tags=tags)
        except TypeError as e:
            error_msg = str(e)
            if "got an unexpected keyword argument" in error_msg:
                raise TypeError(
                    f"{error_msg}\n\nHint: You may need to update your task function signature. "
                    f"Either add the missing parameter to your function definition, or use "
                    f"**kwargs to accept any additional parameters."
                ) from e
            raise

        if task_result is None:
            return None

        if isinstance(task_result, TaskResult):
            return task_result

        if isinstance(task_result, str):
            return TaskResult(output=task_result, metadata=None, tags=None)

        raise TypeError(
            f"task returned unexpected unexpected type {type(task_result)!r}. "
            f"Allowed types: {TaskResult.__name__!r}, 'str' and 'NoneType'."
        )

    async def evaluate_stage(
        self,
        adapted_evaluators: list[BaseEvaluatorAdapter],
        row: datasets.Row,
        task_result: TaskResult,
        parent: EvalParent,
        tags: Tags,
    ) -> list[EvaluationResult]:
        attrs = {"tags": tags, "experiment_tags": self.tags, "dataset_id": row.dataset_id, "dataset_sample_id": row.sid}
        with evaluation_attributes(attrs=attrs):
            evals_gen = (
                with_semaphore(self._sem_evals, adapter.evaluate(row, task_result, parent))
                for adapter in adapted_evaluators
            )

            if len(adapted_evaluators) > 1:
                with bundled_eval("Evaluations"):
                    results = await asyncio.gather(*evals_gen, return_exceptions=True)
            else:
                results = await asyncio.gather(*evals_gen, return_exceptions=True)
        return results


async def with_semaphore(sem, coro):
    async with sem:
        return await coro


def generate_experiment_name(name: str) -> str:
    ts = int(time.time())
    if name:
        return f"{name}-{ts}"
    try:
        login = os.getlogin()
        return f"{login}-{ts}"
    except OSError:  # Possible in-cluster error: No such device or address
        return str(ts)


def _adapt_evaluators(evaluators: list[AdaptableEvaluators]) -> list[BaseEvaluatorAdapter]:
    def into(e):
        if isinstance(e, BaseEvaluatorAdapter):
            return e
        return StructuredEvaluatorAdapter(e)

    return [into(e) for e in evaluators]


def _trace_task(task):
    if task is None:
        return None
    if hasattr(task, "_pat_traced"):
        return task
    return traced(attributes={"gen_ai.operation.name": "task"})(task)



import abc
import contextlib
import contextvars
import datetime
import functools
import inspect
import logging
import threading
import time
import typing
import uuid
from opentelemetry.trace import get_current_span, SpanContext
from typing import Any, Optional, Union

from patronus import context
from patronus.api.api_client import PatronusAPIClient
from patronus.evals.context import get_context_evaluation_attributes
from patronus.api import api_types
from patronus.evals.types import EvaluationResult
from patronus.exceptions import UninitializedError
from patronus.retry import retry
from patronus.tracing.attributes import LogTypes, GenAIAttributes, OperationNames
from patronus.tracing.decorators import start_span
from patronus.tracing.logger import Logger
from patronus.utils import merge_tags

log = logging.getLogger("patronus.core")


LogID = uuid.UUID


class EvaluationDataRecord(typing.NamedTuple):
    """
    EvaluationDataRecord holds evaluation log data and ID of the log.
    Log data consists of positional arguments (args) and keyword arguments (kwargs)
    that were passed to an evaluator.
    """

    log_id: LogID
    arguments: dict

    def __str__(self):
        return str({"log_id": self.log_id, "arguments": self.arguments})


class UniqueEvaluationDataSet:
    """
    UniqueEvaluationDataSet holds unique list evaluation log data.
    This container is used to track and ensure that duplicated evaluation data logs are not emitted.
    """

    span_context: SpanContext
    logs: typing.List[EvaluationDataRecord]
    lock: threading.Lock

    __slots__ = ("span_context", "logs", "lock")

    def __init__(self, span_context: SpanContext):
        self.span_context = span_context
        self.logs = []
        self.lock = threading.Lock()

    def __str__(self):
        return str({"id": id(self), "span_context": self.span_context, "logs": self.logs})

    def __enter__(self) -> "UniqueEvaluationDataSet":
        self.lock.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()

    def find_log(self, bound_arguments) -> Optional[LogID]:
        """Must be used with lock"""
        for lg in self.logs:
            if bound_arguments == lg.arguments:
                return lg.log_id
        return None

    def _add_log(self, arguments: dict[str, Any]) -> EvaluationDataRecord:
        """Must be used with lock"""
        eval_log = EvaluationDataRecord(
            log_id=uuid.uuid4(),
            arguments=arguments,
        )
        self.logs = [*self.logs, eval_log]
        return eval_log

    def log(
        self, *, logger: Logger, is_method: bool, self_key_name: Optional[str], bound_arguments: dict[str, Any]
    ) -> LogID:
        if is_method:
            assert self_key_name is not None
            bound_arguments = {**bound_arguments}
            bound_arguments.pop(self_key_name)
        with self.lock:
            log_id = self.find_log(bound_arguments)
            if log_id is not None:
                return log_id
            record = self._add_log(bound_arguments)
        logger.log(
            body=record.arguments,
            log_type=LogTypes.eval,
            log_id=record.log_id,
            span_context=self.span_context,
        )
        return record.log_id


_ctx_evaluation_log_group: contextvars.ContextVar[UniqueEvaluationDataSet] = contextvars.ContextVar("_ctx_bundled_eval")


def get_current_log_id(bound_arguments: dict[str, Any]) -> Optional[LogID]:
    """
    Return log_id for given arguments in current context.
    Returns None if there is no context - most likely SDK is not initialized.
    """
    eval_group = _ctx_evaluation_log_group.get(None)
    if eval_group is None:
        return None
    log_id = eval_group.find_log(bound_arguments)
    if log_id is None:
        raise ValueError("Log not found for provided arguments")
    return log_id


@contextlib.contextmanager
def _start_evaluation_log_group() -> typing.Iterator[UniqueEvaluationDataSet]:
    span = get_current_span()
    value = UniqueEvaluationDataSet(span_context=span.get_span_context())
    tokens = _ctx_evaluation_log_group.set(value)
    yield value
    _ctx_evaluation_log_group.reset(tokens)


@contextlib.contextmanager
def _get_or_start_evaluation_log_group() -> typing.Iterator[UniqueEvaluationDataSet]:
    ctx = _ctx_evaluation_log_group.get(None)
    if ctx is not None:
        yield ctx
    else:
        with _start_evaluation_log_group() as ctx:
            yield ctx


@contextlib.contextmanager
def bundled_eval(span_name: str = "Evaluation bundle"):
    tracer = context.get_tracer_or_none()
    if tracer is None:
        yield
        return

    with tracer.start_as_current_span(span_name):
        with _start_evaluation_log_group():
            yield


def handle_eval_output(
    *,
    ctx: context.PatronusContext,
    log_id: LogID,
    evaluator_id: str,
    criteria: Optional[str] = None,
    metric_name: str,
    metric_description: str,
    ret_value: typing.Any,
    duration: datetime.timedelta,
    qualname: str,
):
    # Returned None means no evaluation took place.
    if ret_value is None:
        return

    ev: EvaluationResult = coerce_eval_output_type(ret_value, qualname)

    span_context = get_current_span().get_span_context()
    evaluation_attrs = get_context_evaluation_attributes()
    tags = merge_tags(evaluation_attrs["tags"], ev.tags, evaluation_attrs["experiment_tags"])
    try:
        eval_payload = api_types.ClientEvaluation(
            log_id=log_id,
            project_name=ctx.scope.project_name,
            app=ctx.scope.app,
            experiment_id=ctx.scope.experiment_id,
            evaluator_id=evaluator_id,
            criteria=criteria,
            pass_=ev.pass_,
            score=ev.score,
            text_output=ev.text_output,
            metadata=ev.metadata,
            explanation=ev.explanation,
            explanation_duration=ev.explanation_duration,
            evaluation_duration=ev.evaluation_duration or duration,
            metric_name=metric_name,
            metric_description=metric_description,
            dataset_id=evaluation_attrs["dataset_id"],
            dataset_sample_id=evaluation_attrs["dataset_sample_id"],
            created_at=datetime.datetime.now(datetime.timezone.utc),
            tags=tags,
            trace_id=hex(span_context.trace_id)[2:].zfill(32),
            span_id=hex(span_context.span_id)[2:].zfill(16),
        )
        ctx.exporter.submit(eval_payload)
    except Exception:
        log.exception("Failed to submit evaluation payload to the exported")

    # # Debug evaluation log
    # ev_data = sorted([f"{k}={v!r}" for k, v in ev.model_dump().items() if v is not None])
    # get_logger().warning(f"Evaluation({', '.join(ev_data)})")


def coerce_eval_output_type(ev_output: typing.Any, qualname: str) -> EvaluationResult:
    if isinstance(ev_output, EvaluationResult):
        return ev_output
    if isinstance(ev_output, bool):
        return EvaluationResult(pass_=ev_output, score=float(ev_output))
    if isinstance(ev_output, (int, float)):
        return EvaluationResult(score=float(ev_output))
    if isinstance(ev_output, str):
        return EvaluationResult(text_output=ev_output)

    raise TypeError(
        f"Evaluator '{qualname}' returned unexpected type {type(ev_output)!r}. "
        f"Supported return types are EvaluationResult, int, float, bool, str. "
    )


class PrepEval(typing.NamedTuple):
    evaluator_id: str
    criteria: Optional[str]
    metric_name: str
    metric_description: Optional[str]
    self_key_name: Optional[str]
    arguments: dict[str, typing.Any]
    disable_export: bool

    def display_name(self):
        if not self.criteria:
            return self.evaluator_id
        return f"{self.criteria} ({self.evaluator_id})"


def _as_applied_argument(signature: inspect.Signature, bound_arguments: inspect.BoundArguments):
    """
    This function is very similar to inspect.BoundArguments.apply_default().

    The difference is that it's not meant to be used for arguments bounding but for as a log record of evaluation data.
    The purpose of it is to provide nicer display for arguments passed to evaluation functions.
    This function will ignore empty varied positional arguments (*args) and varied keyword arguments (**kwargs)
    if are empty (were not provided by the function called).
    """
    arguments = bound_arguments.arguments
    new_arguments = []
    for name, param in signature.parameters.items():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            for kwname, kwvalue in arguments.get(name, {}).items():
                new_arguments.append((kwname, kwvalue))
        else:
            try:
                new_arguments.append((name, arguments[name]))
            except KeyError:
                if param.default is not inspect.Parameter.empty:
                    val = param.default
                else:
                    continue
                new_arguments.append((name, val))
    return dict(new_arguments)


def evaluator(
    *,
    # Name for the evaluator. Defaults to function name (or class name in case of class based evaluators).
    evaluator_id: Union[str, typing.Callable[[], str], None] = None,
    # Name of the criteria used by the evaluator.
    # The use of the criteria is only recommended in more complex evaluator setups
    # where evaluation algorithm changes depending on a criteria (think strategy pattern).
    criteria: Union[str, typing.Callable[[], str], None] = None,
    # Name for the evaluation metric. Defaults to evaluator_id value.
    metric_name: Optional[str] = None,
    # The description of the metric used for evaluation.
    # If not provided then the docstring of the wrapped function is used for this value.
    metric_description: Optional[str] = None,
    # Whether the wrapped function is a method.
    # This value is used to determine whether to remove "self" argument from the log.
    # It also allows for dynamic evaluator_id and criteria discovery
    # based on `get_evaluator_id()` and `get_criteria_id()` methods.
    # User-code usually shouldn't use it as long as user defined class-based evaluators inherit from
    # the library provided Evaluator base classes.
    is_method: bool = False,
    **kwargs,
):
    def decorator(fn):
        fn_sign = inspect.signature(fn)

        def _get_eval_id():
            return (callable(evaluator_id) and evaluator_id()) or evaluator_id or fn.__qualname__

        def _get_criteria():
            return (callable(criteria) and criteria()) or criteria or None

        def _prep(*fn_args, **fn_kwargs):
            bound_args = fn_sign.bind(*fn_args, **fn_kwargs)
            arguments_to_log = _as_applied_argument(fn_sign, bound_args)
            bound_args.apply_defaults()
            self_key_name = None
            instance = None
            if is_method:
                self_key_name = next(iter(fn_sign.parameters.keys()))
                instance = bound_args.arguments[self_key_name]

            eval_id = None
            eval_criteria = None
            if isinstance(instance, Evaluator):
                eval_id = instance.get_evaluator_id()
                eval_criteria = instance.get_criteria()

            if eval_id is None:
                eval_id = _get_eval_id()
            if eval_criteria is None:
                eval_criteria = _get_criteria()

            met_name = metric_name or eval_id
            met_description = metric_description or inspect.getdoc(fn) or None

            disable_export = isinstance(instance, RemoteEvaluatorMixin) and instance._disable_export

            return PrepEval(
                evaluator_id=eval_id,
                criteria=eval_criteria,
                metric_name=met_name,
                metric_description=met_description,
                self_key_name=self_key_name,
                arguments=arguments_to_log,
                disable_export=disable_export,
            )

        @functools.wraps(fn)
        async def wrapper_async(*fn_args, **fn_kwargs):
            ctx = context.get_current_context_or_none()
            if ctx is None:
                return await fn(*fn_args, **fn_kwargs)

            prep = _prep(*fn_args, **fn_kwargs)

            start = time.perf_counter()
            try:
                with start_span(prep.display_name(), attributes={GenAIAttributes.operation_name: OperationNames.eval}):
                    with _get_or_start_evaluation_log_group() as log_group:
                        log_id = log_group.log(
                            logger=ctx.pat_logger,
                            is_method=is_method,
                            self_key_name=prep.self_key_name,
                            bound_arguments=prep.arguments,
                        )
                        ret = await fn(*fn_args, **fn_kwargs)
            except Exception as e:
                ctx.logger.exception(f"Evaluator raised an exception: {e}")
                raise e
            if prep.disable_export:
                return ret
            elapsed = time.perf_counter() - start
            handle_eval_output(
                ctx=ctx,
                log_id=log_id,
                evaluator_id=prep.evaluator_id,
                criteria=prep.criteria,
                metric_name=prep.metric_name,
                metric_description=prep.metric_description,
                ret_value=ret,
                duration=datetime.timedelta(seconds=elapsed),
                qualname=fn.__qualname__,
            )
            return ret

        @functools.wraps(fn)
        def wrapper_sync(*fn_args, **fn_kwargs):
            ctx = context.get_current_context_or_none()
            if ctx is None:
                return fn(*fn_args, **fn_kwargs)

            prep = _prep(*fn_args, **fn_kwargs)

            start = time.perf_counter()
            try:
                with start_span(prep.display_name(), attributes={GenAIAttributes.operation_name: OperationNames.eval}):
                    with _get_or_start_evaluation_log_group() as log_group:
                        log_id = log_group.log(
                            logger=ctx.pat_logger,
                            is_method=is_method,
                            self_key_name=prep.self_key_name,
                            bound_arguments=prep.arguments,
                        )
                        ret = fn(*fn_args, **fn_kwargs)
            except Exception as e:
                ctx.logger.exception("Evaluation failed")
                raise e
            if prep.disable_export:
                return ret
            elapsed = time.perf_counter() - start
            handle_eval_output(
                ctx=ctx,
                log_id=log_id,
                evaluator_id=prep.evaluator_id,
                criteria=prep.criteria,
                metric_name=prep.metric_name,
                metric_description=prep.metric_description,
                ret_value=ret,
                duration=datetime.timedelta(seconds=elapsed),
                qualname=fn.__qualname__,
            )
            return ret

        def _set_attrs(wrapper: Any):
            wrapper._pat_evaluator = True

            # _pat_evaluator_id and _pat_criteria_id may be a bit misleading since
            # may not be correct since actually values for evaluator_id and criteria
            # are dynamically dispatched for class-based evaluators.
            # These values will be correct for function evaluators though.
            wrapper._pat_evaluator_id = _get_eval_id()
            wrapper._pat_criteria = _get_criteria()

        if inspect.iscoroutinefunction(fn):
            _set_attrs(wrapper_async)
            return wrapper_async
        else:
            _set_attrs(wrapper_sync)
            return wrapper_sync

    return decorator


# Inherit from ABCMeta so we can enforce abstractmethods
class _EvaluatorMeta(abc.ABCMeta):
    def __new__(mcls, name, bases, namespace, /, **kwargs):
        if (
            "evaluate" in namespace
            and
            # Skip wrapping in case the class is still an abstract class
            getattr(namespace["evaluate"], "__isabstractmethod__", None) is not True
        ):
            # Wrap evaluate method with evaluator decorator as soon as class is created.
            # This happens on class creation, not class instantiation.
            evaluator_id = namespace.get("evaluator_id", name)
            into_evaluator = evaluator(
                evaluator_id=evaluator_id,
                criteria=namespace.get("criteria"),
                is_method=True,
            )
            namespace["evaluate"] = into_evaluator(namespace["evaluate"])
        return super().__new__(mcls, name, bases, namespace, **kwargs)


class Evaluator(metaclass=_EvaluatorMeta):
    evaluator_id: Optional[str] = None
    criteria: Optional[str] = None

    def get_evaluator_id(self) -> str:
        return self.evaluator_id or self.__class__.__qualname__

    def get_criteria(self) -> Optional[str]:
        return self.criteria

    @abc.abstractmethod
    def evaluate(self, *args, **kwargs) -> Optional[EvaluationResult]:
        """
        Synchronous version of evaluate method.
        When inheriting directly from Evaluator class it's permitted to change parameters signature.
        Return type should stay unchanged.
        """


class AsyncEvaluator(Evaluator):
    @abc.abstractmethod
    async def evaluate(self, *args, **kwargs) -> Optional[EvaluationResult]:
        """
        Asynchronous version of evaluate method.
        When inheriting directly from Evaluator class it's permitted to change parameters signature.
        Return type should stay unchanged.
        """


class StructuredEvaluator(Evaluator):
    @abc.abstractmethod
    def evaluate(
        self,
        *,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_attachments: Union[list[Any], None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[typing.Dict[str, typing.Any]] = None,
        **kwargs: Any,
    ) -> EvaluationResult: ...


class AsyncStructuredEvaluator(AsyncEvaluator):
    @abc.abstractmethod
    async def evaluate(
        self,
        *,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_attachments: Union[list[Any], None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[typing.Dict[str, typing.Any]] = None,
        **kwargs: Any,
    ) -> EvaluationResult: ...


class RemoteEvaluatorMixin:
    _disable_export = True

    def __init__(
        self,
        evaluator_id_or_alias: str,
        criteria: Optional[str] = None,
        *,
        tags: Optional[dict[str, str]] = None,
        explain_strategy: typing.Literal["never", "on-fail", "on-success", "always"] = "always",
        criteria_config: Optional[dict[str, typing.Any]] = None,
        allow_update: bool = False,
        max_attempts: int = 3,
        api_: Optional[PatronusAPIClient] = None,
    ):
        self.evaluator_id_or_alias = evaluator_id_or_alias
        self.criteria = criteria
        self.tags = tags or {}
        self.explain_strategy = explain_strategy
        self.criteria_config = criteria_config
        self.allow_update = allow_update
        self.max_attempts = max_attempts
        self._api = api_

    def get_evaluator_id(self) -> str:
        return self.evaluator_id_or_alias

    def get_criteria(self) -> str:
        return self.criteria

    def _get_api(self) -> PatronusAPIClient:
        api_client = self._api or context.get_api_client_or_none()
        if api_client is None:
            raise UninitializedError(
                "Failed to get API client. Either initialize the library by calling patronus.init() or "
                "Apss the API client object to the RemoveEvaluator constructor."
            )
        return api_client

    @staticmethod
    def _translate_response(resp: api_types.EvaluationResult) -> EvaluationResult:
        return EvaluationResult(
            score=resp.score_raw,
            pass_=resp.pass_,
            text_output=resp.text_output,
            metadata=resp.additional_info,
            explanation=resp.explanation,
            tags=resp.tags,
            evaluation_duration=resp.evaluation_duration,
            explanation_duration=resp.explanation_duration,
        )


class RemoteEvaluator(RemoteEvaluatorMixin, StructuredEvaluator):
    def evaluate(
        self,
        *,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_attachments: Union[list[Any], None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[typing.Dict[str, typing.Any]] = None,
        **kwargs: Any,
    ) -> EvaluationResult:
        kws = {
            "system_prompt": system_prompt,
            "task_context": task_context,
            "task_attachments": task_attachments,
            "task_input": task_input,
            "task_output": task_output,
            "gold_answer": gold_answer,
            "task_metadata": task_metadata,
            **kwargs,
        }
        log_id = get_current_log_id(bound_arguments=kws)

        attrs = get_context_evaluation_attributes()
        tags = {**self.tags}
        if t := attrs["tags"]:
            tags.update(t)
        tags = merge_tags(tags, kwargs.get("tags"), attrs["experiment_tags"])
        if tags:
            kws["tags"] = tags
        if did := attrs["dataset_id"]:
            kws["dataset_id"] = did
        if sid := attrs["dataset_sample_id"]:
            kws["dataset_sample_id"] = sid

        resp = retry()(self._evaluate)(log_id=log_id, **kws)
        return self._translate_response(resp)

    def _evaluate(
        self,
        *,
        log_id: Optional[LogID] = None,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_attachments: Union[list[Any], None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[typing.Dict[str, typing.Any]] = None,
        **kwargs: Any,
    ) -> api_types.EvaluationResult:
        project_name = None
        app = None
        experiment_id = None
        ctx = context.get_current_context_or_none()
        if ctx is not None:
            project_name = ctx.scope.project_name
            app = ctx.scope.app
            experiment_id = ctx.scope.experiment_id

        span_context = get_current_span().get_span_context()
        return self._get_api().evaluate_one_sync(
            api_types.EvaluateRequest(
                evaluators=[
                    api_types.EvaluateEvaluator(
                        evaluator=self.evaluator_id_or_alias,
                        criteria=self.criteria,
                        explain_strategy=self.explain_strategy,
                    )
                ],
                evaluated_model_system_prompt=system_prompt,
                evaluated_model_retrieved_context=task_context,
                evaluated_model_input=task_input,
                evaluated_model_output=task_output,
                evaluated_model_gold_answer=gold_answer,
                evaluated_model_attachments=task_attachments,
                project_id=None,
                project_name=project_name,
                app=app,
                experiment_id=experiment_id,
                capture="all",
                dataset_id=kwargs.get("dataset_id"),
                dataset_sample_id=kwargs.get("dataset_sample_id"),
                tags=kwargs.get("tags"),
                trace_id=hex(span_context.trace_id)[2:].zfill(32),
                span_id=hex(span_context.span_id)[2:].zfill(16),
                log_id=log_id and str(log_id),
            )
        )


class AsyncRemoteEvaluator(RemoteEvaluatorMixin, AsyncStructuredEvaluator):
    async def evaluate(
        self,
        *,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_attachments: Union[list[Any], None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[typing.Dict[str, typing.Any]] = None,
        **kwargs: Any,
    ) -> EvaluationResult:
        kws = {
            "system_prompt": system_prompt,
            "task_context": task_context,
            "task_attachments": task_attachments,
            "task_input": task_input,
            "task_output": task_output,
            "gold_answer": gold_answer,
            "task_metadata": task_metadata,
            **kwargs,
        }
        log_id = get_current_log_id(bound_arguments=kws)

        attrs = get_context_evaluation_attributes()
        tags = {**self.tags}
        if t := attrs["tags"]:
            tags.update(t)
        tags = merge_tags(tags, kwargs.get("tags"), attrs["experiment_tags"])
        if tags:
            kws["tags"] = tags
        if did := attrs["dataset_id"]:
            kws["dataset_id"] = did
        if sid := attrs["dataset_sample_id"]:
            kws["dataset_sample_id"] = sid

        resp = await retry()(self._evaluate)(log_id=log_id, **kws)
        return self._translate_response(resp)

    async def _evaluate(
        self,
        *,
        log_id: Optional[LogID] = None,
        system_prompt: Optional[str] = None,
        task_context: Union[list[str], str, None] = None,
        task_attachments: Union[list[Any], None] = None,
        task_input: Optional[str] = None,
        task_output: Optional[str] = None,
        gold_answer: Optional[str] = None,
        task_metadata: Optional[typing.Dict[str, typing.Any]] = None,
        **kwargs: Any,
    ) -> api_types.EvaluationResult:
        project_name = None
        app = None
        experiment_id = None
        ctx = context.get_current_context_or_none()
        if ctx is not None:
            project_name = ctx.scope.project_name
            app = ctx.scope.app
            experiment_id = ctx.scope.experiment_id

        span_context = get_current_span().get_span_context()
        return await self._get_api().evaluate_one(
            api_types.EvaluateRequest(
                evaluators=[
                    api_types.EvaluateEvaluator(
                        evaluator=self.evaluator_id_or_alias,
                        criteria=self.criteria,
                        explain_strategy=self.explain_strategy,
                    )
                ],
                evaluated_model_system_prompt=system_prompt,
                evaluated_model_retrieved_context=task_context,
                evaluated_model_input=task_input,
                evaluated_model_output=task_output,
                evaluated_model_gold_answer=gold_answer,
                evaluated_model_attachments=task_attachments,
                project_id=None,
                project_name=project_name,
                app=app,
                experiment_id=experiment_id,
                capture="all",
                dataset_id=kwargs.get("dataset_id"),
                dataset_sample_id=kwargs.get("dataset_sample_id"),
                tags=kwargs.get("tags"),
                trace_id=hex(span_context.trace_id)[2:].zfill(32),
                span_id=hex(span_context.span_id)[2:].zfill(16),
                log_id=log_id and str(log_id),
            )
        )
